#include "Analysis.hpp"
#include <algorithm>

using namespace std;

void Analysis::estimateElimCosts(Node *node)
{
	uint64_t i;
	uint64_t j;

	uint64_t totalDofs;

	totalDofs = 0;
	for (uint64_t supernode : node->getSupernodes()) {
		totalDofs += (*node->meshSupernodes)[supernode];
	}

	j = 0;
	for (i = 0; i < node->getSupernodesToElim(); ++i) {
		uint64_t dofs = (*node->meshSupernodes)[node->getSupernodes()[i]];
		fprintf(stdout, "%lu = %lu (%d %d)\n",
			node->getSupernodes()[i],
			node->getFLOPs(dofs, totalDofs - j), dofs, totalDofs-j);
		j += dofs;
	}

	if (node->getLeft() && node->getRight()) {
		Analysis::estimateElimCosts(node->getLeft());
		Analysis::estimateElimCosts(node->getRight());
	}

}

void Analysis::nodeAnaliser(Node *node, set<uint64_t> *parent)
{
	auto getAllSupernodes = [] (Node *n) {
		set<uint64_t> *supernodes = new set<uint64_t>();
		for (Element *e : n->getElements()) {
			for (uint64_t supernode : e->supernodes)
				supernodes->insert(supernode);
		}
		return supernodes;
	};

	set<uint64_t> *common;

	if (node->getLeft() != NULL && node->getRight() != NULL) {
		set<uint64_t> *lSupernodes = getAllSupernodes(node->getLeft());
		set<uint64_t> *rSupernodes = getAllSupernodes(node->getRight());

		common = new set<uint64_t>;
		std::set_intersection(lSupernodes->begin(), lSupernodes->end(),
				      rSupernodes->begin(), rSupernodes->end(),
				      std::inserter(*common, common->begin()));

		delete (lSupernodes);
		delete (rSupernodes);

		for (auto p = parent->cbegin(); p!=parent->cend(); ++p) {
			common->insert(*p);
		}

		Analysis::nodeAnaliser(node->getLeft(), common);
		Analysis::nodeAnaliser(node->getRight(), common);

	} else {
		common = getAllSupernodes(node);
	}

	int i = 0;

	for (uint64_t supernode : *common) {
		if (!parent->count(supernode)) {
			node->addSupernode(supernode);
			++i;
		}
	}

	node->setSupernodesToElim(i);

	for (uint64_t supernode : *common) {
		if (parent->count(supernode)) {
			node->addSupernode(supernode);
		}
	}

	delete common;
}

void Analysis::doAnalise(Mesh *mesh)
{
	Node *root = mesh->getRootNode();
	std::set<uint64_t> *parent = new set<uint64_t>();
	Analysis::nodeAnaliser(root, parent);
	Analysis::mergeAnaliser(root);
	delete parent;
}

void Analysis::mergeAnaliser(Node *node)
{
	Node *left;
	Node *right;
	int i;

	left = node->getLeft();
	right = node->getRight();

	if (left != NULL && right != NULL) {
		node->leftPlaces.resize(left->getSupernodes().size() - left->getSupernodesToElim());
		node->rightPlaces.resize(right->getSupernodes().size() - right->getSupernodesToElim());

		map<uint64_t, uint64_t> reverseMap;

		for (i=0; i<node->getSupernodes().size(); ++i) {
			reverseMap[node->getSupernodes()[i]] = i;
		}

		for (i=left->getSupernodesToElim(); i<left->getSupernodes().size(); ++i) {
			node->leftPlaces[i-left->getSupernodesToElim()] = reverseMap[left->getSupernodes()[i]];
		}

		for (i=right->getSupernodesToElim(); i<right->getSupernodes().size(); ++i) {
			node->rightPlaces[i-right->getSupernodesToElim()] = reverseMap[right->getSupernodes()[i]];
		}

		Analysis::mergeAnaliser(left);
		Analysis::mergeAnaliser(right);
	}
	node->computeOffsets();
}


