c----------------------------------------------------------------------
c
c   module name        - elimination tree routines
c
c----------------------------------------------------------------------
c
c   computer           - machine independent
c
c   latest revision    - June 09
c
c   purpose            - contains elimination tree data structure 
c                        with some methods working on the 
c                        elimination tree
c
c----------------------------------------------------------------------
      
      module elimination_tree_mod

c      use supernodes_system_mod

      type tree_node
c.......unique id (corresponds to the node number in elimination tree)
        integer :: id
c.......tree node connectivities
        integer :: father
        integer :: nr_sons
        integer, dimension(:),pointer :: sons

c.......tree node stores list of subdomains (blocks)
        integer :: nr_subdomains
        integer, dimension(:), pointer :: subdomains
      end type tree_node

      type tree3
c
        type(tree_node),dimension(:),allocatable :: TREE_NODES
        integer :: NR_TREE_NODES
        integer :: MAX_TREE_NODES
c
        integer :: root
        integer :: new_root
      end type tree3

      type(tree3) :: TREE_INSTANCE
c
      allocatable :: ielems_at_node(:,:) !used for dumpint out the tree
c

      contains
c----------------------------------------------------------------------
      subroutine reset_tree
      NR_TREE_NODES=0
      MAX_TREE_NODES=0
      root=0
      new_root=0      
      end subroutine reset_tree
c----------------------------------------------------------------------
c  ...creates new tree
      subroutine create_tree(nr_nodes)
c
      iprint=0
c
      TREE_INSTANCE%MAX_TREE_NODES=nr_nodes
      allocate(TREE_INSTANCE%TREE_NODES(
     .  TREE_INSTANCE%MAX_TREE_NODES),STAT=i1)
      if(i1/=0)then
        write(*,*)'create_tree: allocation error',i1,nr_nodes
        stop
      endif
      if(iprint.eq.1)then
        write(*,*)
     .  'Allocation TREE_NODES(',TREE_INSTANCE%MAX_TREE_NODES,')'
      endif
c-> out of core
      do i=1,nr_nodes
        TREE_INSTANCE%TREE_NODES(i)%id=i
      enddo
c<-
      TREE_INSTANCE%NR_TREE_NODES=0
      end subroutine create_tree
c----------------------------------------------------------------------
c  ...creates tree root
      subroutine create_root(subdomains,nr_subdomains)
      integer, dimension(:) ::subdomains
c
      iprint=0
c
      TREE_INSTANCE%NR_TREE_NODES = 1
      TREE_INSTANCE%TREE_NODES(1)%nr_sons=0
      TREE_INSTANCE%TREE_NODES(1)%father=0
c
      if(iprint.eq.1)then
        write(*,*)'create_root:nr_subdomains',nr_subdomains
      endif
      TREE_INSTANCE%TREE_NODES(1)%nr_subdomains = nr_subdomains
      allocate(TREE_INSTANCE%TREE_NODES(1)%subdomains(nr_subdomains),
     .  STAT=i1)
      if(i1/=0)then
        write(*,*)'create_root: allocation error',i1
        stop
      endif
      TREE_INSTANCE%TREE_NODES(1)%subdomains(1:nr_subdomains) = 
     .   subdomains(1:nr_subdomains)
      TREE_INSTANCE%root = 1
c
      if(iprint.eq.1)then
        write(*,*)'exiting create_tree'
      endif
      end subroutine create_root
c----------------------------------------------------------------------
c  ...creates tree root
      subroutine create_new_root(subdomains,nr_subdomains)
      integer, dimension(:) ::subdomains
c
      iprint=1
c
      TREE_INSTANCE%NR_TREE_NODES = 
     .   TREE_INSTANCE%NR_TREE_NODES+1
      TREE_INSTANCE%TREE_NODES(
     .   TREE_INSTANCE%NR_TREE_NODES)%nr_sons=0
      TREE_INSTANCE%TREE_NODES(
     .   TREE_INSTANCE%NR_TREE_NODES)%father=0
c
      if(iprint.eq.1)then
        write(*,*)'create_root:nr_subdomains',nr_subdomains
      endif
      TREE_INSTANCE%TREE_NODES(
     .  TREE_INSTANCE%NR_TREE_NODES)%nr_subdomains = nr_subdomains
      allocate(TREE_INSTANCE%TREE_NODES(
     .  TREE_INSTANCE%NR_TREE_NODES)%subdomains(nr_subdomains),
     .  STAT=i1)
      if(i1/=0)then
        write(*,*)'create_root: allocation error',i1
        stop
      endif
      TREE_INSTANCE%TREE_NODES(
     .  TREE_INSTANCE%NR_TREE_NODES)%subdomains(1:nr_subdomains) = 
     .   subdomains(1:nr_subdomains)
c
      if(iprint.eq.1)then
        write(*,*)'exiting create_tree'
      endif
      end subroutine create_new_root
c----------------------------------------------------------------------
c  ...destroys tree
      subroutine delete_tree
c
      call delete_node_rec(TREE_INSTANCE%root)
      deallocate(TREE_INSTANCE%TREE_NODES)
      TREE_INSTANCE%NR_TREE_NODES=0
      TREE_INSTANCE%MAX_TREE_NODES=0
      end subroutine delete_tree
c- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
      recursive subroutine delete_node_rec(node)
      integer :: node, iprint
c
      iprint=0
c
      if(TREE_INSTANCE%TREE_NODES(node)%nr_sons.gt.0)then
        if(iprint.eq.1)then
          write(*,*)PRINTRANK,'delete_node_rec:nr_sons',
     .    TREE_INSTANCE%TREE_NODES(node)%nr_sons
        endif
        do ison=1,TREE_INSTANCE%TREE_NODES(node)%nr_sons
         if(iprint.eq.1)then
           write(*,*)PRINTRANK,'call delete_node_rec for son',ison
         endif
         call delete_node_rec(TREE_INSTANCE%TREE_NODES(node)%sons(ison))
         if(iprint.eq.1)then
           write(*,*)PRINTRANK,'----------after call for son',ison
         endif
        enddo
        deallocate(TREE_INSTANCE%TREE_NODES(node)%sons)
      endif
      if(iprint.eq.1)then
        write(*,*)PRINTRANK,'delete_node_rec:call deallocate_system'
      endif
      if(associated(TREE_INSTANCE%TREE_NODES(node)%subdomains))then
        deallocate(TREE_INSTANCE%TREE_NODES(node)%subdomains)
      endif
      nullify(TREE_INSTANCE%TREE_NODES(node)%subdomains)
      end subroutine delete_node_rec
c----------------------------------------------------------------------
      subroutine create_new_son(father,nr_subdomains,subdomains)
c  ...pointer to father node
      integer :: father
c  ...number of subdomains assigned to son node
      integer :: nr_subdomains
c  ...pointer to subdomains list
      integer, dimension(:),allocatable :: subdomains
c.....local data
      integer,dimension(:),pointer :: father_sons
c
      iprint=0
c
      if(iprint.eq.1)then
        write(*,*)'entering create_new_son',father
      endif
c
      if(TREE_INSTANCE%TREE_NODES(father)%nr_sons.gt.0)then
        allocate(father_sons(TREE_INSTANCE%TREE_NODES(father)%nr_sons),
     .    STAT=i1)
        if(i1/=0)then
          write(*,*)'create_new_son: allocation error',i1
          stop
        endif
        father_sons(1:TREE_INSTANCE%TREE_NODES(father)%nr_sons)=
     .    TREE_INSTANCE%TREE_NODES(father)%sons(
     .       1:TREE_INSTANCE%TREE_NODES(father)%nr_sons)
        deallocate(TREE_INSTANCE%TREE_NODES(father)%sons)
        allocate(TREE_INSTANCE%TREE_NODES(father)%sons(
     .    TREE_INSTANCE%TREE_NODES(father)%nr_sons+1),STAT=i1)
        if(i1/=0)then
          write(*,*)'create_new_son: allocation error',i1
          stop
        endif
        do ison=1,TREE_INSTANCE%TREE_NODES(father)%nr_sons
           TREE_INSTANCE%TREE_NODES(father)%sons(ison)=father_sons(ison)
        enddo
        deallocate(father_sons)
      else
        allocate(TREE_INSTANCE%TREE_NODES(father)%sons(1),STAT=i1)
        if(i1/=0)then
          write(*,*)'create_new_son: allocation error',i1
          stop
        endif
      endif
      if(iprint.eq.1)then
        write(*,*)'adding new son number',
     .  'TREE_INSTANCE%TREE_NODES(',father,')%nr_sons+1'
      endif
c.....add a new node
      TREE_INSTANCE%TREE_NODES(father)%nr_sons = 
     .  TREE_INSTANCE%TREE_NODES(father)%nr_sons+1
      TREE_INSTANCE%NR_TREE_NODES = TREE_INSTANCE%NR_TREE_NODES+1
c
      TREE_INSTANCE%TREE_NODES(father)%sons(
     .  TREE_INSTANCE%TREE_NODES(father)%nr_sons) = 
     .    TREE_INSTANCE%NR_TREE_NODES

c.....initialize the son node
      if(iprint.eq.1)then
        write(*,*)'initialize the son node'
        write(*,*)'step1'
      endif
      TREE_INSTANCE%TREE_NODES(TREE_INSTANCE%NR_TREE_NODES)%nr_sons=0
      if(iprint.eq.1)then
        write(*,*)'step2'
      endif
      TREE_INSTANCE%TREE_NODES(TREE_INSTANCE%NR_TREE_NODES)%father = 
     .  father
      if(iprint.eq.1)then
        write(*,*)'step3'
      endif
      TREE_INSTANCE%TREE_NODES(TREE_INSTANCE%NR_TREE_NODES)
     .  %nr_subdomains=nr_subdomains
      if(iprint.eq.1)then
       write(*,*)PRINTRANK,'step4','subdomains(1:',nr_subdomains,')=',
     .  subdomains(1:nr_subdomains)
      endif
      allocate(TREE_INSTANCE%TREE_NODES(
     .  TREE_INSTANCE%NR_TREE_NODES)%subdomains(nr_subdomains),STAT=i1)
      if(i1/=0)then
        write(*,*)'create_new_son: allocation error',i1
        stop
      endif
      TREE_INSTANCE%TREE_NODES(TREE_INSTANCE%NR_TREE_NODES)
     .  %subdomains(1:nr_subdomains) = subdomains(1:nr_subdomains)
c
      if(iprint.eq.1)then
        write(*,*)'TREE_INSTANCE%TREE_NODES(',father,')%nr_sons = ',
     .    TREE_INSTANCE%TREE_NODES(father)%nr_sons
        write(*,*)'exiting create_new_son'
      endif
      end subroutine create_new_son
c----------------------------------------------------------------------
      subroutine create_new_node(nr_subdomains,subdomains,
     .                           nr_neig_list,neig_list,iarea)
c  ...number of subdomains assigned to son node
      integer :: nr_subdomains
c  ...pointer to subdomains list
      integer, dimension(:),allocatable :: subdomains
c  ...number of neig
      integer :: nr_neig_list
c  ...pointer to processors list
      integer, dimension(:),allocatable :: neig_list
c.....local data
      integer,dimension(:),pointer :: father_sons
c.....area of elements
      integer :: iarea
c
      iprint=0
c
      if(iprint.eq.1)then
        write(*,*)'entering create_new_node'
      endif
c
      TREE_INSTANCE%NR_TREE_NODES = TREE_INSTANCE%NR_TREE_NODES+1
c.....initialize the node
      if(iprint.eq.1)then
        write(*,*)'initialize the node'
        write(*,*)'step1'
      endif
      TREE_INSTANCE%TREE_NODES(TREE_INSTANCE%NR_TREE_NODES)%nr_sons=0
      if(iprint.eq.1)then
        write(*,*)'step2'
      endif
      TREE_INSTANCE%TREE_NODES(TREE_INSTANCE%NR_TREE_NODES)%father = 
     .  0
      if(iprint.eq.1)then
        write(*,*)'step3'
      endif
      TREE_INSTANCE%TREE_NODES(TREE_INSTANCE%NR_TREE_NODES)
     .  %nr_subdomains=nr_subdomains
      if(iprint.eq.1)then
       write(*,*)PRINTRANK,'step4','subdomains(1:',nr_subdomains,')=',
     .  subdomains(1:nr_subdomains)
      endif
      allocate(TREE_INSTANCE%TREE_NODES(
     .  TREE_INSTANCE%NR_TREE_NODES)%subdomains(nr_subdomains),STAT=i1)
      if(i1/=0)then
        write(*,*)'create_new_son: allocation error',i1
        stop
      endif
      TREE_INSTANCE%TREE_NODES(TREE_INSTANCE%NR_TREE_NODES)
     .  %subdomains(1:nr_subdomains) = subdomains(1:nr_subdomains)
      if(iprint.eq.1)then
       write(*,*)PRINTRANK,'step5','neig_list(1:',nr_neig_list,')=',
     .  neig_list(1:nr_neig_list)
      endif
c
      if(iprint.eq.1)then
        write(*,*)'exiting create_new_node'
      endif
      end subroutine create_new_node

c----------------------------------------------------------------------
c  ...write down tree data
      subroutine dumpout_tree
      write(40,*)TREE_INSTANCE%NR_TREE_NODES
      call dumpout_node(TREE_INSTANCE%root)
      end subroutine dumpout_tree
c----------------------------------------------------------------------
c  ...write down node
      recursive subroutine dumpout_node(node)
      type(tree3) :: T
      integer :: node
      integer, allocatable, dimension(:,:) :: pair_subdomains
c
      iprint=0
c
      allocate(pair_subdomains(
     .  TREE_INSTANCE%TREE_NODES(node)%nr_subdomains,2))
      do i=1,
     .  TREE_INSTANCE%TREE_NODES(node)%nr_subdomains
        isub=
     .    TREE_INSTANCE%TREE_NODES(node)%subdomains(i)
c        pair_subdomains(i,1)=ielems_at_node(isub+1,1)
c        pair_subdomains(i,2)=ielems_at_node(isub+1,2)
        pair_subdomains(i,1)=1
        pair_subdomains(i,2)=isub+1
      enddo
      if(iprint.eq.1)then
        write(40,*)'--------------------------------------------------'
        write(40,*)'node id',
     .    TREE_INSTANCE%TREE_NODES(node)%id
        write(40,*)'nr_elements',
     .  TREE_INSTANCE%TREE_NODES(node)%nr_subdomains
        write(40,*)'levels,elements:',(pair_subdomains(i,1:2),
     .  i=1,TREE_INSTANCE%TREE_NODES(node)%nr_subdomains)
        if(TREE_INSTANCE%TREE_NODES(node)%nr_sons.eq.2)then
          write(40,*)'sons',
     .    TREE_INSTANCE%TREE_NODES(
     .    TREE_INSTANCE%TREE_NODES(node)%sons(1))%id,
     .    TREE_INSTANCE%TREE_NODES(
     .    TREE_INSTANCE%TREE_NODES(node)%sons(2))%id
        endif
      else
c node id, number of elements, elements (we call them subdomains)
        if(TREE_INSTANCE%TREE_NODES(node)%nr_sons.eq.2)then
          write(40,*)TREE_INSTANCE%TREE_NODES(node)%id,
     .    TREE_INSTANCE%TREE_NODES(node)%nr_subdomains,
     .    (pair_subdomains(i,1:2),i=1,
     .     TREE_INSTANCE%TREE_NODES(node)%nr_subdomains),
     .    TREE_INSTANCE%TREE_NODES(
     .    TREE_INSTANCE%TREE_NODES(node)%sons(1))%id,
     .    TREE_INSTANCE%TREE_NODES(
     .    TREE_INSTANCE%TREE_NODES(node)%sons(2))%id
        else
          write(40,*)TREE_INSTANCE%TREE_NODES(node)%id,
     .    TREE_INSTANCE%TREE_NODES(node)%nr_subdomains,
     .    (pair_subdomains(i,1:2),i=1,
     .     TREE_INSTANCE%TREE_NODES(node)%nr_subdomains)
        endif
      endif
      deallocate(pair_subdomains)
      do ison=1,TREE_INSTANCE%TREE_NODES(node)%nr_sons
        call dumpout_node(TREE_INSTANCE%TREE_NODES(node)%sons(ison))
      enddo
      end subroutine dumpout_node

c----------------------------------------------------------------------      
c  ...write down tree data
      subroutine report_tree
      call report_node(TREE_INSTANCE%root)      
      end subroutine report_tree
c----------------------------------------------------------------------      
c  ...write down node
      recursive subroutine report_node(node)
      type(tree3) :: T
      integer :: node
c
      iprint=1
c
      if(iprint.eq.1)then
        write(*,*)'--------------------------------------------------'
        write(*,*)'Node',node,'father',
     .    TREE_INSTANCE%TREE_NODES(node)%father
        write(*,*)'nr_subdomains',
     .  TREE_INSTANCE%TREE_NODES(node)%nr_subdomains
       write(*,*)'subdomains',TREE_INSTANCE%TREE_NODES(node)%subdomains(
     .  1:TREE_INSTANCE%TREE_NODES(node)%nr_subdomains)
        write(*,*)'nr_sons',TREE_INSTANCE%TREE_NODES(node)%nr_sons
      endif
      do ison=1,TREE_INSTANCE%TREE_NODES(node)%nr_sons
        if(iprint.eq.1)then
          write(*,*)'son:',ison, '-> Node:',
     .    TREE_INSTANCE%TREE_NODES(node)%sons(ison)
        endif
        call report_node(TREE_INSTANCE%TREE_NODES(node)%sons(ison))
      enddo
      call pause
      end subroutine report_node
c----------------------------------------------------------------------
      recursive subroutine built_tree(node,subdomains,nr_subdomains)
c  ...arguments
      integer :: node
      integer, dimension(:) ::subdomains
c  ...local variables
      integer, dimension(:),allocatable :: subdomains_left
      integer, dimension(:),allocatable :: subdomains_right
      integer :: son1,son2
c
      iprint=0
c
      if(nr_subdomains.gt.1)then

        nr_subdomains_half = nr_subdomains / 2
        nr_subdomains_left = nr_subdomains_half
        allocate(subdomains_left(nr_subdomains_left),STAT=i1)
        if(i1/=0)then
          write(*,*)'built_tree: allocation error',i1
          stop
        endif
        do isub=1,nr_subdomains_half
           subdomains_left(isub)=subdomains(isub)
        enddo
        nr_subdomains_right = nr_subdomains - nr_subdomains_left
        allocate(subdomains_right(nr_subdomains_right),STAT=i1)
        if(i1/=0)then
          write(*,*)'built_tree: allocation error',i1
          stop
        endif
        do isub=nr_subdomains_half+1,nr_subdomains
           subdomains_right(isub-nr_subdomains_half)=subdomains(isub)
        enddo
c
        if(iprint.eq.1)then
          write(*,*)'call create_new_son with nr_subdomains',
     .      'subdomains_left(1:',nr_subdomains_left,')=',
     .      subdomains_left(1:nr_subdomains_left)
        endif
        call create_new_son(node,nr_subdomains_left,subdomains_left)
        if(iprint.eq.1)then
          write(*,*)'call create_new_son with subdomains',
     .      'subdomains_right(1:',nr_subdomains_right,')=',
     .      subdomains_right(1:nr_subdomains_right)
        endif
        call create_new_son(node,nr_subdomains_right,subdomains_right)
c
        son1 = TREE_INSTANCE%TREE_NODES(node)%sons(1)
        call built_tree(son1,subdomains_left,nr_subdomains_left)
        son2 = TREE_INSTANCE%TREE_NODES(node)%sons(2)
        call built_tree(son2,subdomains_right,nr_subdomains_right)

        deallocate(subdomains_left)
        deallocate(subdomains_right)
      endif
c
      end subroutine built_tree
c----------------------------------------------------------------------
      end module elimination_tree_mod
