      program main
c    
      use projection_engine
      use partition_engine
      use graphics_engine
c
      implicit none
c      
      integer(kind=4) :: nx, px, nelemx
      real   (kind=8), allocatable, dimension(:) :: Ux
      integer(kind=4) :: ny, py, nelemy
      real   (kind=8), allocatable, dimension(:) :: Uy
      integer(kind=4) :: nz, pz, nelemz
      real   (kind=8), allocatable, dimension(:) :: Uz
      real   (kind=8), allocatable, dimension(:) :: WEx
      real   (kind=8), allocatable, dimension(:) :: WEy
      real   (kind=8), allocatable, dimension(:) :: WEz

      double precision, allocatable, dimension(:,:) :: M
      double precision, allocatable, dimension(:) :: F
      
      integer(kind=4) :: resx,resy,resz,d
      real   (kind=8), allocatable, dimension(:,:) :: F3
      

      integer :: i
      integer :: iret
      integer :: iprint !debug flag
c
      iprint=1
c
      resx=10
      resy=10
      resz=10
      d=1
c  ...prepare the problem dimensions
      px=2 !order
      nx=3 !intervals+1
      allocate(Ux(nx+px+2)) !knot vector
      py=2 !order
      ny=3 !intervals+1
      allocate(Uy(ny+py+2)) !knot vector
      pz=2 !order
      nz=3 !intervals+1
      allocate(Uz(nz+pz+2)) !knot vector
c
      Ux(1:px+1)=0.d0
      Ux(nx+2:nx+px+2)=1.d0
      do i=px+2,nx+1
        Ux(i)=real(i-px-1)/real(nx-px+1)
      enddo
      nelemx = CountSpans(nx,px,Ux)
      if(iprint.eq.1)then
        write(*,*)'nx,px,nelemx',nx,px,nelemx
        write(*,*)'Ux',Ux
      endif
      allocate(WEx(nelemx+1))
      call ComputeWeights(nx,px,nelemx,Ux,WEx)
      if(iprint.eq.1)then
        write(*,*)'WEx',WEx
      endif
c
      Uy(1:py+1)=0.d0
      Uy(ny+2:ny+py+2)=1.d0
      do i=py+2,ny+1
        Uy(i)=real(i-py-1)/real(ny-py+1)
      enddo
      nelemy = CountSpans(ny,py,Uy)
      if(iprint.eq.1)then
        write(*,*)'ny,py,nelemy',ny,py,nelemy
        write(*,*)'Uy',Uy
      endif
      allocate(WEy(nelemy+1))
      call ComputeWeights(ny,py,nelemy,Uy,WEy)
      if(iprint.eq.1)then
        write(*,*)'WEy',WEy
      endif
c
c
      Uz(1:pz+1)=0.d0
      Uz(nz+2:nz+pz+2)=1.d0
      do i=pz+2,nz+1
        Uz(i)=real(i-pz-1)/real(nz-pz+1)
      enddo
      nelemz = CountSpans(nz,pz,Uz)
      if(iprint.eq.1)then
        write(*,*)'nz,pz,nelemz',nz,pz,nelemz
        write(*,*)'Uz',Uz
      endif
      allocate(WEz(nelemz+1))
      call ComputeWeights(nz,pz,nelemz,Uz,WEz)
      if(iprint.eq.1)then
        write(*,*)'WEz',WEz
      endif
c
      call test
     . (Ux,px,nx,nelemx,Uy,py,ny,nelemy,Uz,pz,nz,nelemz)

c
c  ...call recursive bissections
c
      if(iprint.eq.1)then
        write(*,*)'call construct_elimination_tree'
      endif
      call construct_elimination_tree
     . (Ux,px,nx,nelemx,Uy,py,ny,nelemy,Uz,pz,nz,nelemz,WEx,WEy,WEz)

      if(iprint.eq.1)then
        write(*,*)'call dumpout_nodes_and_tree'
      endif
      call dumpout_nodes_and_tree
     . (Ux,px,nx,nelemx,Uy,py,ny,nelemy,Uz,pz,nz,nelemz)

c
c  ...call mesh estimator
c
      call SYSTEM ( " ./meshestimator/analyser -f ./files/tree 
     .  > ./files/analysed " )
      
      allocate(F3((nz+1),(nx+1)*(ny+1))) !z,x,y
      
      d=1
      
      call readout_nodes_and_costs
     . (Ux,px,nx,nelemx,Uy,py,ny,nelemy,Uz,pz,nz,nelemz,F3)
      call Graphics(1,d,
     .   Ux,px,nx,nelemx,resx,
     .   Uy,py,ny,nelemy,resy,
     .   Uz,pz,nz,nelemz,resz,
     .   F3,2)
     
c
      deallocate(F3)
      deallocate(Ux)
      deallocate(Uy)
      deallocate(Uz)
      deallocate(WEx)
      deallocate(WEy)
      deallocate(WEz)
c
      end

