      module partition_engine
      use elimination_tree_mod
      use projection_engine

      integer, allocatable, dimension(:) :: ELEM_NEW_ORDER
      integer, allocatable, dimension(:) :: INVERSE_MAP

      contains


      subroutine global2local(IGlobal,I_n,J_n,K_n,I,J,K)
c
c  ...debug flag
      iprint=0
      if(iprint.eq.1)write(*,*)'I_n,J_n,K_n',I_n,J_n,K_n
      ig=IGlobal-1
      if(iprint.eq.1)write(*,*)'ig',ig      
      K = iabs(ig / ((I_n)*(J_n)))
      if(iprint.eq.1)write(*,*)'K',K,'ig',ig      
      ig=ig-K*(I_n)*(J_n)
      if(iprint.eq.1)write(*,*)'J',J,'ig',ig
      J = ig/(I_n)
      ig=ig-(I_n)*J
      I=ig
      if(iprint.eq.1)write(*,*)'I',I,'ig',ig

c add 1 since local numbers also start from 1      
      I=I+1;J=J+1;K=K+1
      if(iprint.eq.1)then
        write(*,*)'global2local: IGlobal',IGlobal,'->',I,J,K
      endif
c
      call local2global(I,J,K,I_n,J_n,K_n,IGlobalTest)
      if(IGlobal.ne.IGlobalTest)then
        write(*,*)'global->local: ERROR',IGlobal,IGlobalTest
        stop
      endif

      end subroutine global2local

      subroutine local2global(I,J,K,I_n,J_n,K_n,IGlobal)
c
      IGlobal = I + (J-1)*(I_n)+(K-1)*(I_n)*(J_n)
c
      end subroutine local2global
c
      subroutine neig_mdle(iel,nelemx,nelemy,nelemz,
     .                     WEx,WEy,WEz,
     .                     weights_mdle,nr_weights,
     .                     neigs_mdle,nr_neigs_mdle)
      integer :: mdle,nr_neigs_mdle,nr_weights_mdle
      integer :: neigs_mdle(27),weights_mdle(27)
      real   (kind=8), intent(in) :: WEx(1:nelemx)
      real   (kind=8), intent(in) :: WEy(1:nelemy)
      real   (kind=8), intent(in) :: WEz(1:nelemz)
      integer :: I,J,K, iprint
      iprint=0
      if(iprint.eq.1)then
        write(*,*)'iel',iel
        write(*,*)'nelemx,nelemy,nelemz',nelemx,nelemy,nelemz
      endif
      call global2local(iel,nelemx,nelemy,nelemz,I,J,K)        
      nr_neigs_mdle=0; nr_weights=0
      if(.not.(I.eq.nelemx))then
         nr_neigs_mdle=nr_neigs_mdle+1
         nr_weights=nr_weights+1
         neigs_mdle(nr_neigs_mdle)=iel+1 !along X+
         weights_mdle(nr_weights)=WEx(I+1)
      endif
      if(.not.(I.eq.1))then
         nr_neigs_mdle=nr_neigs_mdle+1
         nr_weights=nr_weights+1
         neigs_mdle(nr_neigs_mdle)=iel-1 !along X-
         weights_mdle(nr_weights)=WEx(I)
      endif
      if(.not.(J.eq.nelemy))then
         nr_neigs_mdle=nr_neigs_mdle+1
         nr_weights=nr_weights+1
         neigs_mdle(nr_neigs_mdle)=iel+nelemx !along Y+
         weights_mdle(nr_weights)=WEy(J+1)
      endif
      if(.not.(J.eq.1))then
         nr_neigs_mdle=nr_neigs_mdle+1
         nr_weights=nr_weights+1
         neigs_mdle(nr_neigs_mdle)=iel-nelemx !along Y-
         weights_mdle(nr_weights)=WEy(J)
      endif
      if(.not.(K.eq.nelemz))then
         nr_neigs_mdle=nr_neigs_mdle+1
         nr_weights=nr_weights+1
         neigs_mdle(nr_neigs_mdle)=iel+nelemx*nelemy !along Z+
         weights_mdle(nr_weights)=WEz(K+1)
      endif
      if(.not.(K.eq.1))then
         nr_neigs_mdle=nr_neigs_mdle+1
         nr_weights=nr_weights+1
         neigs_mdle(nr_neigs_mdle)=iel-nelemx*nelemy !along Z-
         weights_mdle(nr_weights)=WEz(K)
      endif
      end subroutine neig_mdle
c
c--------------------------------------------------------------------------------------------
      subroutine construct_elimination_tree
     .    (Ux,px,nx,nelemx,Uy,py,ny,nelemy,Uz,pz,nz,nelemz,WEx,WEy,WEz)
c
c.....arguments
      integer :: ibeg, iend,ilevel, node_tree
      integer(kind=4), intent(in) :: nx, px, nelemx
      real   (kind=8), intent(in) :: Ux(0:nx+px+1)
      integer(kind=4), intent(in) :: ny, py, nelemy
      real   (kind=8), intent(in) :: Uy(0:ny+py+1)
      integer(kind=4), intent(in) :: nz, pz, nelemz
      real   (kind=8), intent(in) :: Uz(0:nz+pz+1)
      real   (kind=8), intent(in) :: WEx(1:nelemx)
      real   (kind=8), intent(in) :: WEy(1:nelemy)
      real   (kind=8), intent(in) :: WEz(1:nelemz)
c
      integer, dimension(:),allocatable ::subdomains
c
      iprint = 0
c
      allocate(ELEM_NEW_ORDER(nelemx*nelemy*nelemz))
      iel=1  
      do i=1,nelemx
         do j=1,nelemy
           do k=1,nelemz
             ELEM_NEW_ORDER(iel)=iel
             iel=iel+1
           enddo
         enddo
      enddo
      maxmdle = nelemx*nelemy*nelemz
      allocate(INVERSE_MAP(maxmdle))
      INVERSE_MAP(1:maxmdle)=0
      do iel=1,maxmdle
         mdle=ELEM_NEW_ORDER(iel)
         INVERSE_MAP(mdle)=iel
      enddo
c
c-> ELIMINATION TREE
      nr_subdomains = maxmdle 
c
      allocate(subdomains(nr_subdomains))
c
      do isub=1,nr_subdomains
        subdomains(isub)=isub-1 !subdomains start from 0
      enddo
c
      call create_tree(nr_subdomains*2) 
      call create_root(subdomains,nr_subdomains)
      deallocate(subdomains)

c<- ELIMINATION TREE
c
      nlevel=1
      nbeg=1
      nend=maxmdle
      nproc_beg=0
      nproc_end=0
      call partition_elements_into_two_with_metis
     .   (nbeg,nend,nlevel,TREE_INSTANCE%root,
     .    Ux,px,nx,nelemx,Uy,py,ny,nelemy,Uz,pz,nz,nelemz,WEx,WEy,WEz)
c
      if(iprint.eq.1)then
        call report_tree()
        call pause
      endif
      deallocate(INVERSE_MAP)
      deallocate(ELEM_NEW_ORDER)      
c
      end subroutine construct_elimination_tree
     
c-----------------------------------------------------------------------
      recursive subroutine partition_elements_into_two_with_metis
     .   (ibeg, iend, ilevel, node_tree, 
     .    Ux,px,nx,nelemx,Uy,py,ny,nelemy,Uz,pz,nz,nelemz,
     .    WEx,WEy,WEz)
c
c.....arguments
      integer :: ibeg, iend,ilevel, node_tree
      integer(kind=4), intent(in) :: nx, px, nelemx
      real   (kind=8), intent(in) :: Ux(0:nx+px+1)
      integer(kind=4), intent(in) :: ny, py, nelemy
      real   (kind=8), intent(in) :: Uy(0:ny+py+1)
      integer(kind=4), intent(in) :: nz, pz, nelemz
      real   (kind=8), intent(in) :: Uz(0:nz+pz+1)
      real   (kind=8), intent(in) :: WEx(1:nelemx)
      real   (kind=8), intent(in) :: WEy(1:nelemy)
      real   (kind=8), intent(in) :: WEz(1:nelemz)
c....local variables
      integer, allocatable, dimension(:) :: subdomains_left
      integer, allocatable, dimension(:) :: subdomains_right
      integer :: nr_subdomains_left, nr_subdomains_right
      integer, allocatable, dimension(:) :: processors_left
      integer, allocatable, dimension(:) :: processors_right
      integer :: nr_processors_left, nr_processors_right
c
      integer  :: mdle, nr_neig_mdle,nr_weight_mdle,iel
      integer :: neigs_mdle(27),weights_mdle(27)
c
c  ...graph for METIS for storing element connecitivities
      integer, allocatable, dimension(:) :: xadj
      integer, allocatable, dimension(:) :: adjncy
      integer, allocatable, dimension(:) :: vwgt
      integer, allocatable, dimension(:) :: adjwgt
      integer, allocatable, dimension(:) :: nperm
      integer, allocatable, dimension(:) :: ioptions
      integer, allocatable, dimension(:) :: subdomains
      integer, allocatable, dimension(:) :: temp_elems
      real, allocatable, dimension(:) :: tpwgts
c
      iprint = 0
c
      do iel=ibeg,iend
         mdle=ELEM_NEW_ORDER(iel)
         INVERSE_MAP(mdle)=iel-ibeg+1
      enddo
c
      nr_elements = iend-ibeg+1 !elements in the tree node
c
      write(*,*)ilevel,'-partition_elements_into_two_with_metis:'
      write(*,*)'subdomains ',ibeg,iend
c
c  ...compute number of neighbors of elements (edges in graph)
c
c
      nr_neigs=0
      do iel=ibeg,iend
        mdle = ELEM_NEW_ORDER(iel)
        call neig_mdle(mdle,nelemx,nelemy,nelemz,
     .                 WEx,WEy,WEz,
     .                 weights_mdle,nr_weights_mdle,
     .                 neigs_mdle,nr_neigs_mdle)
        if(iprint.eq.1)then
          write(*,*)'neig_mdle for',mdle,'returns'
          write(*,*)'neigs_mdle',neigs_mdle(1:nr_neigs_mdle)
          write(*,*)'weights_mdle',weights_mdle(1:nr_weights_mdle)
        endif
        do i=1,nr_neigs_mdle
c  .......additionally, check if the neighbor is located
c         within the graph being partitioned
          ifound=0
          do iels=ibeg,iend
             if(neigs_mdle(i).eq.ELEM_NEW_ORDER(iels))ifound=1
          enddo
          if(ifound.eq.0)cycle
          if(iprint.eq.1)write(*,*)'add neig',neigs_mdle(i)
          nr_neigs=nr_neigs+1
        enddo
        if(iprint.eq.1)write(*,*)'nr_neigs',nr_neigs
      enddo
c
      if(nr_neigs.eq.0)then
        if(iprint.eq.1)then
          write(*,*)'completely disconnected graph'
        endif
         ihalf = (ibeg+iend)/2
         go to 10
c        return
      endif

      if(iprint.eq.1)then
        write(*,*)'>>>>>>>>>>>>>>>>>>>nr_neigs',nr_neigs
      endif
c
c it is assumed that elements are number level by level, x, y and z
c

c--------------------------------------------------------------------
c  ...allocate the graph data structures
      allocate(xadj(nr_elements+1),STAT=i1) !location of 1st neighbor
      allocate(adjncy(nr_neigs),STAT=i2) !on the adjacency list
c     it is assumed that nodes are number from 1 (fortran style)
      allocate(vwgt(nr_elements),STAT=i3)
      allocate(adjwgt(nr_neigs),STAT=i4)
      if(i1+i2+i3+i4/=0)then
        write(*,*)'partition_elements_into_two_with_metis:',
     .    ' graph allocation arror',i1,i2,i3,i4
        stop
      endif
c--------------------------------------------------------------------
c  ...fill the graph data structure
      neig_location=0
      do iel=ibeg,iend
        xadj(iel-ibeg+1)=neig_location
        vwgt(iel-ibeg+1)=1
c
        mdle = ELEM_NEW_ORDER(iel)
        call neig_mdle(mdle,nelemx,nelemy,nelemz,
     .                 WEx,WEy,WEz,
     .                 weights_mdle,nr_weights_mdle,
     .                 neigs_mdle,nr_neigs_mdle)
        if(iprint.eq.1)then
          write(*,*)'neig_mdle for',mdle,'returns'
          write(*,*)'neigs_mdle',neigs_mdle(1:nr_neigs_mdle)
          write(*,*)'weights_mdle',weights_mdle(1:nr_weights_mdle)
        endif
        do i=1,nr_neigs_mdle
            mdle_neig = neigs_mdle(i)
c  .........additionally, check if the neighbor is located
c           within the graph being partitioned
            ifound=0
            do iels=ibeg,iend
               if(neigs_mdle(i).eq.ELEM_NEW_ORDER(iels))ifound=1
            enddo
            if(ifound.eq.0)cycle
            neig_location=neig_location+1
            if(neig_location.gt.nr_neigs)then
                write(*,*)'partition_elements_into_two_with_metis:',
     .          'neig_location.gt.nr_neigs',neig_location,nr_neigs
                stop
            endif
            adjncy(neig_location)=INVERSE_MAP(mdle_neig)
            adjwgt(neig_location) = weights_mdle(i)
        enddo
        if(iprint.eq.1)write(*,*)'neig_location',neig_location
      enddo
      xadj(nr_elements+1)=nr_neigs+1

c      write(*,*)'nr_neigs',nr_neigs,'neig_location',neig_location
c      call pause

c--------------------------------------------------------------------
c<- PART*
cc  ...calling METIS
cc     it is assumed that nodes are number from 1 (fortran style)
c      numflag=1
c      iwgtflag=2 !3 edges and vertices !2 vertices only !1edges only !0 no weights
c      nparts=2 !partition into 2 parts
c      allocate(ioptions(5))
c      ioptions(1:5)=0
c      iedgecut=0
c      allocate(nperm(nr_elements),STAT=i2)
c      if(i1+i2/=0)then
c        write(*,*)': nperm allocatin error',i1,i2
c        stop
c      endif
c      if(iprint.eq.1)then
c        write(*,*)'calling metis'
c        write(*,*)'nr_elements',nr_elements
c        write(*,*)'xadj',xadj
c        call pause
c        write(*,*)'adjncy',adjncy
c        write(*,*)'vwgt',vwgt
c        write(*,*)'adjwgt',adjwgt
c        call pause
c      endif
cc      call METIS_PARTGRAPHRECURSIVE(nr_elements,xadj,adjncy,vwgt,adjwgt,
cc     .  iwgtflag,numflag,nparts,ioptions,iedgecut,nperm)
c      call METIS_PARTGRAPHKWAY(nr_elements,xadj,adjncy,vwgt,adjwgt,
c     .  iwgtflag,numflag,nparts,ioptions,iedgecut,nperm)
cc VKWAY crashes
cc      call METIS_PARTGRAPHVKWAY(nr_elements,xadj,adjncy,vwgt,adjwgt,
cc     .  iwgtflag,numflag,nparts,ioptions,iedgecut,nperm)
c<- PART*
c-> WPartGraphRecursive
c     it is assumed that nodes are number from 1 (fortran style)
      numflag=1 !fortran numbering - elements start from 0
      iwgtflag=2 !3 edges and vertices !2 vertices only !edges only !0 nothing
      nparts=2 !partition into 2 parts
      allocate(ioptions(5))
      ioptions(1:5)=0
      iedgecut=0
      allocate(nperm(nr_elements),STAT=i2)
      allocate(tpwgts(2),STAT=i3)
      tpwgts(1:2)=0.5d0
      if(i1+i2+i3/=0)then
        write(*,*)': nperm allocatin error',i1,i2,i3
        stop
      endif
c      iprint=1
      if(iprint.eq.1)then
        write(*,*)'calling metis'
        write(*,*)'nr_elements',nr_elements
        write(*,*)'xadj',xadj
        call pause
        write(*,*)'adjncy',adjncy
        write(*,*)'vwgt',vwgt
        write(*,*)'adjwgt',adjwgt
        write(*,*)'tpwgts',tpwgts
        call pause
      endif
c      call METIS_WPARTGRAPHRECURSIVE(nr_elements,xadj,adjncy,vwgt,
c     .  adjwgt,iwgtflag,numflag,nparts,tpwgts,ioptions,iedgecut,nperm)
c      call METIS_PARTGRAPHKWAY(nr_elements,xadj,adjncy,vwgt,adjwgt,
c     .  iwgtflag,numflag,nparts,ioptions,iedgecut,nperm)
      call METIS_PARTGRAPHRECURSIVE(nr_elements,xadj,adjncy,vwgt,adjwgt,
     .  iwgtflag,numflag,nparts,ioptions,iedgecut,nperm)
      if(iprint.eq.1)then
        write(*,*)'after calling metis'
c        write(*,*)'iedgecut',iedgecut
c        write(*,*)'nperm',nperm
        call pause
      endif
c at this point, we have element subdomains set in nperm(iel) into 1 or 2
c......................................................................
c  ...reorganize ELEM_NEW_ORDER
c
      if(iprint.eq.1)then
        write(*,*)'reorganize ELEM_NEW_ORDER'
      endif
      allocate(temp_elems(nr_elements),STAT=i1)
      if(i1.ne.0)then
        write(*,*)'partition_elements_into_two_with_metis:',
     .    ' temp_elems',i1
        stop
      endif
      n=0
      do iel=ibeg,iend
        if(nperm(iel-ibeg+1).eq.1)then
          n=n+1
          temp_elems(n)=ELEM_NEW_ORDER(iel)
        endif
      enddo
      ihalf = ibeg+n-1
      do iel=ibeg,iend
        if(nperm(iel-ibeg+1).eq.2)then
          n=n+1
          temp_elems(n)=ELEM_NEW_ORDER(iel)
        endif
      enddo
      if(n.ne.nr_elements)then
         write(*,*)'partition_elements_into_two_with_metis:',
     .     'n!=nr_elements',n,nr_elements
      endif
      deallocate(xadj,adjncy,vwgt,adjwgt,ioptions,nperm)
c
      ELEM_NEW_ORDER(ibeg:iend)=temp_elems(1:nr_elements)
      deallocate(temp_elems)
c
 10   continue
c-> ELIMINATION TREE
        if(iprint.eq.1)then
          write(*,*)'construct elemination tree'
        endif
      if(ihalf.eq.ibeg)then
c special case, e.g. ibeg=11, ihalf=11, iend=13
        nr_subdomains_left = 1
        allocate(subdomains_left(nr_subdomains_left))
        subdomains_left(1)=ibeg   -1!since subdomains are counted from 0
        call create_new_son
     .    (node_tree,nr_subdomains_left,subdomains_left)
        deallocate(subdomains_left)
c
        nr_subdomains_right = iend-ihalf
        allocate(subdomains_right(nr_subdomains_right))
        do i=1,nr_subdomains_right
          subdomains_right(i)=ihalf+i  -1!since subdomains are counted from 0
        enddo
        call create_new_son
     .    (node_tree,nr_subdomains_right,subdomains_right)
        deallocate(subdomains_right)
        ison1 = TREE_INSTANCE%TREE_NODES(node_tree)%sons(1)
        ison2 = TREE_INSTANCE%TREE_NODES(node_tree)%sons(2)
      else
        nr_subdomains_left = ihalf-ibeg+1
        allocate(subdomains_left(nr_subdomains_left))
        do i=1,nr_subdomains_left
          subdomains_left(i)=ibeg+i-1   -1!since subdomains are counted from 0
        enddo
        call create_new_son
     .    (node_tree,nr_subdomains_left,subdomains_left)
        deallocate(subdomains_left)
        nr_subdomains_right = iend-ihalf
        allocate(subdomains_right(nr_subdomains_right))
        do i=1,nr_subdomains_right
          subdomains_right(i)=(ihalf+1)+i-1  -1!since subdomains are counted from 0
        enddo
        call create_new_son
     .    (node_tree,nr_subdomains_right,subdomains_right)
        deallocate(subdomains_right)
        ison1 = TREE_INSTANCE%TREE_NODES(node_tree)%sons(1)
        ison2 = TREE_INSTANCE%TREE_NODES(node_tree)%sons(2)
      endif
c<- ELIMINATION TREE
c
      ilevel=ilevel+1

c---> DEBUG
        if(iprint.eq.1)then
          write(*,*)'ibeg,ihalf,iend',ibeg,ihalf,iend
          write(*,*)'node_tree%subdomains',
     .      TREE_INSTANCE%TREE_NODES(node_tree)%subdomains
     .        (1:TREE_INSTANCE%TREE_NODES(node_tree)%nr_subdomains)
          write(*,*)'node_tree%nr_sons',
     .      TREE_INSTANCE%TREE_NODES(node_tree)%nr_sons
          write(*,*)'son1%subdomains',
     .      TREE_INSTANCE%TREE_NODES(ison1)%subdomains
     .        (1:TREE_INSTANCE%TREE_NODES(ison1)%nr_subdomains)
          write(*,*)'son2%subdomains',
     .      TREE_INSTANCE%TREE_NODES(ison2)%subdomains
     .        (1:TREE_INSTANCE%TREE_NODES(ison2)%nr_subdomains)
          call pause
c<--- DEBUG
        endif
c
c      write(*,*)'DEBUG:3 or more subdomains in the first half'
c
c - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
c     3 or more subdomains in the first half
c - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      if(ihalf-ibeg.gt.1)then
c  .....set elements to take part in the partition
c        do iel=ibeg,ihalf
c          iel_order = ELEM_NEW_ORDER(iel)
c          ELEMS(iel_order)%proc_owner=iel-ibeg+1
c        enddo
        nbeg=ibeg; nhalf=ihalf; nlevel=ilevel
        call partition_elements_into_two_with_metis
     .    (nbeg,nhalf,nlevel,ison1,
     .     Ux,px,nx,nelemx,Uy,py,ny,nelemy,Uz,pz,nz,nelemz,WEx,WEy,WEz)
c  .....reset elements to take part in the partition
c        do iel=ibeg,ihalf
c          iel_order = ELEM_NEW_ORDER(iel)
c          ELEMS(iel_order)%proc_owner=0
c        enddo
      endif


c      write(*,*)'DEBUG:2 subdomains in the first half'
c - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
c     2 subdomains in the first half
c - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      if(ihalf-ibeg.eq.1)then
c-> ELIMINATION TREE
        nr_subdomains_left = 1
        allocate(subdomains_left(nr_subdomains_left))
        subdomains_left(1)=ibeg -1 !since subdomains are counted from 0
        call create_new_son
     .    (ison1,nr_subdomains_left,subdomains_left)
        deallocate(subdomains_left)
c
        nr_subdomains_right = 1
        allocate(subdomains_right(nr_subdomains_right))
        subdomains_right(1)=ihalf -1 !since subdomains are counted from 0
        call create_new_son
     .    (ison1,nr_subdomains_right,subdomains_right)
        deallocate(subdomains_right)
c---> DEBUG
        if(iprint.eq.1)then
          write(*,*)'son1%subdomains',
     .      TREE_INSTANCE%TREE_NODES(ison1)%subdomains
     .        (1:TREE_INSTANCE%TREE_NODES(ison1)%nr_subdomains)
          write(*,*)'son1%nr_sons',
     .      TREE_INSTANCE%TREE_NODES(ison1)%nr_sons
          ison1_1 = TREE_INSTANCE%TREE_NODES(ison1)%sons(1)
          write(*,*)'son1%son1%subdomains',
     .      TREE_INSTANCE%TREE_NODES(ison1_1)%subdomains
     .        (1:TREE_INSTANCE%TREE_NODES(ison1_1)%nr_subdomains)
          ison1_2 = TREE_INSTANCE%TREE_NODES(ison1)%sons(2)
          write(*,*)'son1%son2%subdomains',
     .      TREE_INSTANCE%TREE_NODES(ison1_2)%subdomains
     .        (1:TREE_INSTANCE%TREE_NODES(ison1_2)%nr_subdomains)
          call pause
        endif
c<--- DEBUG

c<- ELIMINATION TREE
      endif
c      write(*,*)'DEBUG:3 or more subdomains in the second half'
c - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
c     3 or more subdomains in the second half
c - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      if(iend-(ihalf+1).gt.1)then
c  .....set elements to take part in the partition
c        do iel=ihalf+1,iend
c          iel_order = ELEM_NEW_ORDER(iel)
c          ELEMS(iel_order)%proc_owner=iel-(ihalf+1)+1
c        enddo
        nhalf=ihalf+1; nend=iend; nlevel=ilevel
        call partition_elements_into_two_with_metis
     .    (nhalf,nend,nlevel,ison2,
     .     Ux,px,nx,nelemx,Uy,py,ny,nelemy,Uz,pz,nz,nelemz,WEx,WEy,WEz)
c  .....reset elements to take part in the partition
c        do iel=ihalf+1,iend
c          iel_order = ELEM_NEW_ORDER(iel)
c          ELEMS(iel_order)%proc_owner=0
c        enddo
      endif
c      write(*,*)'DEBUG:2 subdomains in the second half'
c - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
c     2 subdomains in the second half
c - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      if(iend-(ihalf+1).eq.1)then
c-> ELIMINATION TREE
        nr_subdomains_left = 1
        allocate(subdomains_left(nr_subdomains_left))
        subdomains_left(1)=ihalf+1  -1 !since subdomains are counted from 0
        call create_new_son
     .    (ison2,nr_subdomains_left,subdomains_left)
        deallocate(subdomains_left)
c
        nr_subdomains_right = 1
        allocate(subdomains_right(nr_subdomains_right))
        subdomains_right(1)=iend   -1 !since subdomains are counted from 0
        call create_new_son
     .    (ison2,nr_subdomains_right,subdomains_right)
        deallocate(subdomains_right)
c---> DEBUG
        if(iprint.eq.1)then
          write(*,*)'son2%subdomains',
     .      TREE_INSTANCE%TREE_NODES(ison2)%subdomains
     .        (1:TREE_INSTANCE%TREE_NODES(ison2)%nr_subdomains)
          write(*,*)'son2%nr_sons',
     .      TREE_INSTANCE%TREE_NODES(ison2)%nr_sons
          ison2_1 = TREE_INSTANCE%TREE_NODES(ison2)%sons(1)
          write(*,*)'son2%son1%subdomains',
     .      TREE_INSTANCE%TREE_NODES(ison2_1)%subdomains
     .        (1:TREE_INSTANCE%TREE_NODES(ison2_1)%nr_subdomains)
          ison2_2 = TREE_INSTANCE%TREE_NODES(ison2)%sons(2)
          write(*,*)'son2%son2%subdomains',
     .      TREE_INSTANCE%TREE_NODES(ison2_2)%subdomains
     .        (1:TREE_INSTANCE%TREE_NODES(ison2_2)%nr_subdomains)
          call pause
        endif
c<--- DEBUG
c<- ELIMINATION TREE
      endif
c - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
c
      if(iprint.eq.1)then
        write(*,*)ilevel,'-exiting'
        write(*,*)'------------------------'
      endif
c
      end subroutine partition_elements_into_two_with_metis

      subroutine dumpout_nodes_and_tree
     .   (Ux,px,nx,nelemx,Uy,py,ny,nelemy,Uz,pz,nz,nelemz)
      !!$ d is dimension
      !!$ U is the knot vector
      !!$ p is the polynomial order
      !!$ n is the index of the last control point
      !!$ nelem you get from running CountSpans
      implicit none
      integer(kind=4), intent(in) :: nx, px, nelemx
      real   (kind=8), intent(in) :: Ux(0:nx+px+1)
      integer(kind=4), intent(in) :: ny, py, nelemy
      real   (kind=8), intent(in) :: Uy(0:ny+py+1)
      integer(kind=4), intent(in) :: nz, pz, nelemz
      real   (kind=8), intent(in) :: Uz(0:nz+pz+1)
c
      integer(kind=4) :: mx,my,mz,ngx,ngy,ngz,ex,ey,ez
      integer(kind=4) :: ex1,ey1,ez1
      integer(kind=4) :: kx,ky,kz,ax,ay,az,d
      integer(kind=4) :: kx1,ky1,kz1,ax1,ay1,az1
      integer(kind=4) :: Ox(nelemx),Oy(nelemy),Oz(nelemz)
      real   (kind=8) :: Jx(nelemx),Jy(nelemy),Jz(nelemz)
      real   (kind=8) :: Wx(px+1),Wy(py+1),Wz(pz+1)
      real   (kind=8) :: Xx(px+1,nelemx)
      real   (kind=8) :: Xy(py+1,nelemy)
      real   (kind=8) :: Xz(pz+1,nelemz)
      real   (kind=8) :: NNx(0:0,0:px,px+1,nelemx),
     .                   NNy(0:0,0:py,py+1,nelemy),
     .                   NNz(0:0,0:pz,pz+1,nelemz)
      integer(kind=4) :: n, iel, idof, nelem_dofs, ind,  
     .                   iprint, i, nelem, ilayer
      integer(kind=4), allocatable, dimension(:) :: nelement_dofs

      iprint=0 ! for debugging the output
c
      mx  = nx+px+1
      ngx = px+1
      my  = ny+py+1
      ngy = py+1
      mz  = nz+pz+1
      ngz = pz+1
      d = 0
c
      call BasisData(px,mx,Ux,d,ngx,nelemx,Ox,Jx,Wx,Xx,NNx)
      call BasisData(py,my,Uy,d,ngy,nelemy,Oy,Jy,Wy,Xy,NNy)
      call BasisData(pz,mz,Uz,d,ngz,nelemz,Oz,Jz,Wz,Xz,NNz)
c TO DO write Ox,Oy,Oz
c
      open(unit=40,file='files/tree',
     .     form='formatted',access='sequential',status='unknown')

      if(iprint.eq.1)write(40,*)'number of nodes'       
c      n = (nelemx+px-1)*(nelemy+py-1)*(nelemz+pz-1)
      n = (nelemx+px)*(nelemy+py)*(nelemz+pz)
c
      write(40,*)n
c
      if(iprint.eq.1)write(40,*)'node id / number of dofs'
      do i=1,n
         write(40,*)i,1
      enddo

      if(iprint.eq.1)write(40,*)'Element number'
      nelem = nelemx*nelemy*nelemz
      write(40,*)nelem

c  ...no refinements, so all the elements are on the first layer
      do ilayer=1,1
c  .....loop through elements on layer 
        iel=0
        do ex = 1,nelemx
           do ey = 1,nelemy
              do ez = 1,nelemz
                 iel=iel+1 !element count
                 nelem_dofs = (px+1)*(py+1)*(pz+1)
                 idof=0
                 allocate(nelement_dofs(nelem_dofs))
c                 J = Jx(ex)*Jy(ey)*Jz(ez)
c                 do kx = 1,ngx
c                    do ky = 1,ngy
c                       do kz = 1,ngz
c                          W = Wx(kx)*Wy(ky)*Wz(kz)
                          do ax = 0,px
                             do ay = 0,py
                                do az = 0,pz
c global index of dof
c                                 ind1 = (Ox(ex)+ax) !along x
c                                 ind23 = (Oy(ey)+ay) !along y
c                                 ind23 = ind23 + (Oz(ez)+az)*(ny+1) !along z

                                   ind = (Oz(ez)+az)*(ny+1)*(nx+1)
                                   ind = ind + (Oy(ey)+ay)*(nx+1)
                                   ind = ind + (Ox(ex)+ax)
                                   idof=idof+1
                                   nelement_dofs(idof)=ind+1 !MP +1
                                enddo
                             enddo
                          enddo
c                       enddo
c                    enddo
c                 enddo
c       
                 if(iprint.eq.1)then
                   write(40,*)'level,element,number of nodes,nodes'
                 endif
                 write(40,*)ilayer,iel,
     .             nelem_dofs,nelement_dofs(1:nelem_dofs)
                 deallocate(nelement_dofs)
              enddo
           enddo
        enddo
      enddo
c
      if(iprint.eq.1)write(40,*)'ELIMINATION TREE:'
      call dumpout_tree

      close(40)

      end subroutine dumpout_nodes_and_tree

      
      subroutine readout_nodes_and_costs
     .   (Ux,px,nx,nelemx,Uy,py,ny,nelemy,Uz,pz,nz,nelemz,F3)
      !!$ d is dimension
      !!$ U is the knot vector
      !!$ p is the polynomial order
      !!$ n is the index of the last control point
      !!$ nelem you get from running CountSpans
      implicit none
      integer(kind=4), intent(in) :: nx, px, nelemx
      real   (kind=8), intent(in) :: Ux(0:nx+px+1)
      integer(kind=4), intent(in) :: ny, py, nelemy
      real   (kind=8), intent(in) :: Uy(0:ny+py+1)
      integer(kind=4), intent(in) :: nz, pz, nelemz
      real   (kind=8), intent(in) :: Uz(0:nz+pz+1)
      real   (kind=8), intent(inout) :: 
     .            F3(0:(nz+1)-1,0:(nx+1)*(ny+1)-1)
c
      integer(kind=4) :: mx,my,mz,ngx,ngy,ngz,ex,ey,ez
      integer(kind=4) :: ex1,ey1,ez1
      integer(kind=4) :: kx,ky,kz,ax,ay,az,d
      integer(kind=4) :: kx1,ky1,kz1,ax1,ay1,az1
      integer(kind=4) :: Ox(nelemx),Oy(nelemy),Oz(nelemz)
      real   (kind=8) :: Jx(nelemx),Jy(nelemy),Jz(nelemz)
      real   (kind=8) :: Wx(px+1),Wy(py+1),Wz(pz+1)
      real   (kind=8) :: Xx(px+1,nelemx)
      real   (kind=8) :: Xy(py+1,nelemy)
      real   (kind=8) :: Xz(pz+1,nelemz)
      real   (kind=8) :: NNx(0:0,0:px,px+1,nelemx),
     .                   NNy(0:0,0:py,py+1,nelemy),
     .                   NNz(0:0,0:pz,pz+1,nelemz)
      integer(kind=4) :: n, iel, idof, nelem_dofs, ind,  
     .                   iprint, i, nelem, ilayer
      integer(kind=4), allocatable, dimension(:) :: nelement_dofs
      integer(kind=4) :: ndof,flop,elem1,elem2,loop
      character(1)    :: car
      real   (kind=8), allocatable, dimension(:) :: dofs_flops
      real   (kind=8), allocatable, dimension(:,:,:) :: V
      integer(kind=4) :: ix,iy,iz

      iprint=1 ! for debugging the output
c
      mx  = nx+px+1
      ngx = px+1
      my  = ny+py+1
      ngy = py+1
      mz  = nz+pz+1
      ngz = pz+1
      d = 0
c
      call BasisData(px,mx,Ux,d,ngx,nelemx,Ox,Jx,Wx,Xx,NNx)
      call BasisData(py,my,Uy,d,ngy,nelemy,Oy,Jy,Wy,Xy,NNy)
      call BasisData(pz,mz,Uz,d,ngz,nelemz,Oz,Jz,Wz,Xz,NNz)
      
      open(unit=40,file='files/analysed',
     .     form='formatted',access='sequential',status='unknown')

      n = (nelemx+px)*(nelemy+py)*(nelemz+pz)
      if(iprint.eq.1)write(*,*)'number of nodes ' , n    
      
      allocate(dofs_flops(n))
      
      do loop=1,n
        read(40,*) ndof,car,flop,elem1,elem2
        dofs_flops(ndof+1) = flop
        if(iprint.eq.1) write(*,*) ndof,car,flop,elem1,elem2
      end do
      
      if(iprint.eq.1) then
         do loop=1,n
            write(*,*) dofs_flops(loop)
         end do
      end if
      
      allocate (V(nx+1,ny+1,nz+1))
      
      do ix=1,nx+1
         do iy=1,ny+1
            do iz=1,nz+1
               V(ix,iy,iz) = dofs_flops(iz + (iy-1)*(nz+1)
     .             + (ix-1)*(nz+1)*(ny+1))
            end do
         end do
      end do
      
      
      do ix=0,nx
         do iy=0,ny
            do iz=0,nz
               F3(iz,(nx+1)*iy+ix) = V(ix+1,iy+1,iz+1)
            end do
         end do
      end do
      
      close (40)
      deallocate(dofs_flops)
      deallocate(V)
      end subroutine readout_nodes_and_costs
      

      subroutine test
     . (Ux,px,nx,nelemx,Uy,py,ny,nelemy,Uz,pz,nz,nelemz)

      integer :: iprint !debug flag
c DEBUG
      integer :: mx,ngx,my,ngy,mz,ngz,d
      integer(kind=4), intent(in) :: nx, px, nelemx
      real   (kind=8), intent(in) :: Ux(0:nx+px+1)
      integer(kind=4), intent(in) :: ny, py, nelemy
      real   (kind=8), intent(in) :: Uy(0:ny+py+1)
      integer(kind=4), intent(in) :: nz, pz, nelemz
      real   (kind=8), intent(in) :: Uz(0:nz+pz+1)
      real   (kind=8) :: Wx(px+1)
      real   (kind=8) :: Xx(px+1,nelemx)
      real   (kind=8) :: Wy(py+1)
      real   (kind=8) :: Xy(py+1,nelemy)
      real   (kind=8) :: Wz(pz+1)
      real   (kind=8) :: Xz(pz+1,nelemz)
      integer(kind=4) :: Ox(nelemx),Oy(nelemy),Oz(nelemz)
      real   (kind=8) :: Jx(nelemx),Jy(nelemy),Jz(nelemz)
      real   (kind=8) :: NNx(0:0,0:px,px+1,nelemx),
     .                   NNy(0:0,0:py,py+1,nelemy),
     .                   NNz(0:0,0:pz,pz+1,nelemz)
c
c  DEBUG
      mx  = nx+px+1
      ngx = px+1
      my  = ny+py+1
      ngy = py+1
      mz  = nz+pz+1
      ngz = pz+1
      d = 0
c
      call BasisData(px,mx,Ux,d,ngx,nelemx,Ox,Jx,Wx,Xx,NNx)
      call BasisData(py,my,Uy,d,ngy,nelemy,Oy,Jy,Wy,Xy,NNy)
      call BasisData(pz,mz,Uz,d,ngz,nelemz,Oz,Jz,Wz,Xz,NNz)
c
      write(*,*)'Ox',Ox
      write(*,*)'Ox',Oy
      write(*,*)'Ox',Oz

      end subroutine test

      
      end module partition_engine
